var _ = require('lodash')

if(!RegExp.escape){
    RegExp.escape = function(s) {
        // 1. let str be ToString(S).
        // 2. ReturnIfAbrupt(str).
        var str = String(s);
        // 3. Let cpList be a List containing in order the code
        // points as defined in 6.1.4 of str, starting at the first element of str.
        var cpList = str.split('')
        // 4. let cuList be a new List
        var cuList = [];
        // 5. For each code point c in cpList in List order, do:
        cpList.forEach(function(c){
            // i. If c is a SyntaxCharacter then do:
            if("^$\\.*+?()[]{}|".indexOf(c) !== -1){
                // a. Append "\" to cuList.
                cuList.push("\\");
            }
            // Append c to cpList.
            cuList.push(c);
        })
        // 6. Return str.
        return cuList.join("");
    };
}


var _OPERATORS = [
    ['|', function(left, right){return left | right}],
    ['^', function(left, right){return left ^ right}],
    ['&', function(left, right){return left & right}],
    ['>>', function(left, right){return left >> right}],
    ['<<', function(left, right){return left << right}],
    ['-', function(left, right){return left - right}],
    ['+', function(left, right){return left + right}],
    ['%', function(left, right){return left % right}],
    ['/', function(left, right){return left / right}],
    ['*', function(left, right){return left * right}]
]

var _ASSIGN_OPERATORS = []
_OPERATORS.forEach(function(e){_ASSIGN_OPERATORS.push([e[0]+'=', e[1]])})
_ASSIGN_OPERATORS.push(['=', function(cur, right){return right}])


function JSInterpreter(code, objects){
    if(!objects){
        objects = {}
    }
    this.code = code
    this._functions = {}
    this._objects = objects
}

JSInterpreter.prototype.interpret_statement = function(stmt, local_vars, allow_recurtion){
    //console.log("Analizing statement :", stmt)

    if(allow_recurtion==null) allow_recurtion = 100
    if(allow_recurtion<0) throw "Recurtion limit reached"

    var should_abort = false
    stmt = _.trimStart(stmt)
    var stmt_m = stmt.match(/var\s/)
    var expr
    if(stmt_m)
        expr = stmt.substring(stmt_m[0].length, stmt.length)
    else{
        var return_m = stmt.match(/return(?:\s+|$)/)
        if(return_m){
            expr = stmt.substring(return_m[0].length, stmt.length)
            should_abort = true
        }
        else{
            expr = stmt
        }
    }
    var v = this.interpret_expression(expr, local_vars, allow_recurtion)
    return [v, should_abort]
}

JSInterpreter.prototype.interpret_expression = function(expr, local_vars, allow_recurtion){
    //console.log(expr, ':', "Analizing expression - local variables", Object.keys(local_vars))
    //Object.keys(local_vars).forEach(function(name){console.log(name, ":", local_vars[name].toString())})
    var self = this

    expr = _.trim(expr)

    if(expr=='') return null

    //console.log(expr, ':', "Testing subexpr...")
    if(expr.startsWith('(')){
        var parens_count = 0
        var end_index = 0
        expr.match(/[()]/g).forEach(function(m){
            if(m == '(')
                parens_count += 1
            else{
                parens_count -= 1
                end_index = expr.indexOf(')', end_index+1)
                if(parens_count == 0){
                    var sub_expr = expr.substring(1, end_index)
                    var sub_result = self.interpret_expression(sub_expr, local_vars, allow_recurtion)
                    var remaining_expr = expr.substring(end_index+1, expr.length)
                    if(!remaining_expr || remaining_expr=='')
                        return sub_result
                    else
                        expr = JSON.stringify(sub_result) + remaining_expr
                }
            }
        })
    }
    //console.log(expr, ':', "Testing assing operators")
    var val = null
    _ASSIGN_OPERATORS.forEach(function(e){
        var op = e[0],
            opfunc = e[1]

        var re = new RegExp('([a-zA-Z_$][a-zA-Z_$0-9]*)(?:\\\[([^\\\]]+?)\\\])?\\\s*' + RegExp.escape(op) + '(.*)$')
        var m = re.exec(expr)
        if(!m)
            return
        //console.log(expr, ':', "Handling operator", op)
        var right_val = self.interpret_expression(m[3], local_vars, allow_recurtion-1)
        if(m[2]){
            var lvar = local_vars[m[1]]
            var idx = self.interpret_expression(m[2], local_vars, allow_recurtion)
            if(isNaN(idx)) throw "Not a number"
            var cur = lvar[idx]
            val = opfunc(cur, right_val)
            lvar[idx] = val
        }
        else{
            var cur = local_vars[m[1]]
            val = opfunc(cur, right_val)
            local_vars[m[1]] = val
        }
    })
    if(val!=null)
        return val

    //console.log(expr, ':', "Testing if int")
    if(!isNaN(expr))
        return parseInt(expr)

    //console.log(expr, ':', "Testing if variable")
    var var_m = /^(?!if|return|true|false)([a-zA-Z_$][a-zA-Z_$0-9]*)$/.exec(expr)
    if(var_m){
        //console.log("Found local variable", var_m[1])
        return local_vars[var_m[1]]
    }


    //console.log(expr, ':', "Testing JSON parsing")
    var res = null
    try{
        res = JSON.parse(expr)
    }catch(e){}//console.log(e)}

    if(res!=null){
        //console.log(expr, ':', "Parse JSON expression")
        return res
    }

    //console.log(expr, ':', "Testing array access")
    var array_m = /([a-zA-Z_$][a-zA-Z_$0-9]*)\[(.*?)\]$/.exec(expr)
    if(array_m){
        //console.log(expr, ':', "Found array access")
        var variable = array_m[1]
        var obj

        if(local_vars[variable]){
            //console.log("Found local variable", variable)
            obj = local_vars[variable]
        }
        else {
            if (!self._objects[variable])
                self._objects[variable] = self.extract_object(variable)
            obj = self._objects[variable]
        }
        var member = self.interpret_expression(array_m[2], local_vars, allow_recurtion)
        if(isNaN(member))
            throw new Error(expr+" : Wrong member access")
        return obj[member]
    }

    //console.log(expr, ':', "Testing call expression")
    var m = /^([a-zA-Z_$][a-zA-Z_$0-9]*)\.([^(]+)(?:\(+([^()]*)\))?$/.exec(expr)
    if(m){
        var variable = m[1]
        var member = m[2]
        var arg_str = m[3]

        //console.log(expr, ':', 'Found call expression', variable, member, arg_str)

        var obj

        if(local_vars[variable]){
            //console.log("Found local variable", variable)
            obj = local_vars[variable]
        }
        else {
            if (!self._objects[variable])
                self._objects[variable] = self.extract_object(variable)
            obj = self._objects[variable]
        }

        if(!arg_str && member!='split' && member!='join' && member!='reverse' && member!='slice' && member!='splice'){
            if(member == 'length')
                return obj.length
            //console.log("Returning", variable+'['+member+']')
            return obj[member]
        }

        var argvals = []
        if(arg_str!=''){
            arg_str.split(',').forEach(function(elt){
                //console.log(expr, ':', 'Parsing call argument', elt)
                argvals.push(self.interpret_expression(elt, local_vars, allow_recurtion))
            })
        }

        if (member == 'split')
            return obj.split('')
        if (member == 'join')
            return obj.join(argvals[0])
        if (member == 'reverse') {
            return obj.reverse()
        }
        if (member == 'slice')
            return obj.slice(argvals[0])
        if (member == 'splice') {
            obj.splice(argvals[0], argvals[1])
            return obj
        }

        //console.log("Calling", member, '(', argvals.join(', '), ')')
        return obj[member](argvals)
    }

    m = /([a-zA-Z_$][a-zA-Z_$0-9]*)\[(.+)\]$/.exec(expr)
    if(m){
        var val = local_vars[m[1]]
        var idx = self.interpret_expression(m[2], local_vars, allow_recurtion-1)
        return val[idx]
    }

    _OPERATORS.forEach(function(e){
        var op = e[0],
            opfunc = e[1]

        var re = new RegExp('(.+?)' + RegExp.escape(op) + '(.+)')
        m = re.exec(expr)

        if(!m)
            return

        //console.log(expr, ':', "Analizing for operator", op)
        var repx = self.interpret_statement(m[1], local_vars, allow_recurtion-1)
        if(repx[1]) throw "Premature abort of left side"
        var repy = self.interpret_statement(m[2], local_vars, allow_recurtion-1)
        if(repy[1]) throw "Premature abort of right side"

        //console.log(expr, ':', repx, repy, opfunc(repx[0], repy[0]))

        val = opfunc(repx[0], repy[0])
    })
    if(val != null)
        return val

    m = /^([a-zA-Z_$][a-zA-Z_$0-9]*)\(([a-zA-Z0-9_$,]+)\)$/.exec(expr)
    if(m){
        var fname = m[1]
        var argvals = []
        if(m[2]){
            m[2].split(',').forEach(function(arg){
                argvals.push(!isNaN(arg)? parseInt(arg) : local_vars[arg])
            })
        }
        if(!self._functions[fname])
            self._functions[fname] = self.extract_function(fname)

        //console.log(self._functions)

        return self._functions[fname].apply(self, argvals)
    }

    throw "Insupported js expression : " + expr
}

JSInterpreter.prototype.extract_object = function(objname){
    //console.log("Extracting object", objname)
    var self = this

    var obj = {}
    var re = new RegExp('(?:var\\\s+)?' + RegExp.escape(objname) + '\\\s*=\\\s*\\\{\\\s*(([a-zA-Z$0-9]+\\\s*:\\\s*function\\\(.*?\\\)\\\s*\\\{.*?\\\}(?:,\\\s*)?)*)\\\}\\\s*;', 'g')
    var obj_m = re.exec(this.code)
    if(obj_m && obj_m[1]){
        //console.log("Found object", obj_m[1])
        var fields = obj_m[1].split(',')
        var stmt = ""
        fields.forEach(function(field){
            stmt += field
            var field_m = /([a-zA-Z$0-9]+)\s*:\s*function\(([a-z,]+)\){([^}]+)}/.exec(stmt)
            if(field_m){
                var argnames = field_m[2].split(',')
                obj[field_m[1]] = self.build_function(argnames, field_m[3])
                stmt = ""
            }
            else{
                stmt += ','
            }
        })
    }

    return obj
}

JSInterpreter.prototype.extract_function = function(funcname){
    var re = new RegExp("(?:function\\\s+"+funcname+"|[{;,]\\\s*"+funcname+"\\\s*=\\\s*function|var\\\s+"+funcname+"\\\s*=\\\s*function)\\\s*\\\(([^)]*)\\\)\\\s*\\\{([^}]+)\\\}", 'gm')
    var func_m = re.exec(this.code)

    if(!func_m)
        throw "Couldn't find the function "+funcname
    var argnames = func_m[1].split(',')

    return this.build_function(argnames, func_m[2])
}

JSInterpreter.prototype.call_function = function(funcname, args){
    var f = this.extract_function(funcname)
    return f.apply(this, args)
}

JSInterpreter.prototype.build_function = function(argnames, code){
    //console.log("Building function :", code)
    var self = this
    function resf(args){
        //console.log("Execution of", code, "with args", args.join(', '))
        var local_vars = {}
        if(argnames && argnames.length && args && args.length){
            argnames.forEach(function(name, index){
                local_vars[name] = args[index]
            })
        }
        //Object.keys(local_vars).forEach(function(name){console.log(name, ':', local_vars[name].length || local_vars[name])})
        var res
        var abort = false
        if(code.indexOf(';')!=-1){
            code.split(';').forEach(function(stmt){
                //console.log("Analizing function statement", stmt)
                if(!abort)
                    res = self.interpret_statement(stmt, local_vars)
                if(res[1])
                    abort = true
            })
        }
        else{
            res = self.interpret_statement(code, local_vars)
        }

        return res
    }
    return resf
}

module.exports = JSInterpreter
