//Url parsing
var Url = require('url');
//HTTP requests
var http = require('https');
//Async promises
var Q = require('q');
//HTML parsing into a Jquery-like object
var cheerio = require('cheerio');
//Parse string with html tags into a clean text
var htmlToText = require('html-to-text');
//Utility functions (underscore equivalent)
var _ = require('lodash');
//FFMPEG
var ffmpeg = require('fluent-ffmpeg')
//fs
var fs = require('fs')
//path
var Path = require('path')

var CipherExtract = require('./CipherExtract')

//Parse a query string into an object
function _parse_query(query){
	var props = query.split('&')
	var query_obj = {}
	props.forEach(function(elt){
		var key = decodeURIComponent(elt.split('=')[0])
		query_obj[key] = decodeURIComponent(elt.split('=')[1])
	})
	return query_obj
}

//Wrap every values of the video_info object into an array
function _parse_video_info(video_info){
	var rep = {}
	for(var key in video_info){
		rep[key] = [video_info[key]]
	}
	return rep
}

//Build url with the query object
function _build_url(url, query){
	var parsed_url = Url.parse(url)
	parsed_url.query = query
	return Url.format(parsed_url);
}

//Search all meta tags with the property == name (uses a cherrio (jquery-like) object)
function search_meta(elt, name){
    return elt(
        'meta[itemprop="' + name + '"],' +
        'meta[name="' + name + '"],' +
        'meta[property="' + name + '"],' +
        'meta[id="' + name + '"],' +
        'meta[http-equiv="' + name + '"]')
}

//Get ext from mime type
function mimetype2ext(mt){
    if(!mt)
        return null

    var ext = {
        'audio/mp4': 'm4a',
        'audio/mpeg': 'mp3'
    }
    if(ext[mt])
        return ext[mt]

    var split = mt.split('/')
    var res = split[split.length-1].toLowerCase()
    ext = {
        '3gpp': '3gp',
        'smptett+xml': 'tt',
        'srt': 'srt',
        'ttaf+xml': 'dfxp',
        'ttml+xml': 'ttml',
        'vtt': 'vtt',
        'x-flv': 'flv',
        'x-mp4-fragmented': 'mp4',
        'x-ms-wmv': 'wmv',
        'mpegurl': 'm3u8',
        'x-mpegurl': 'm3u8',
        'vnd.apple.mpegurl': 'm3u8',
        'dash+xml': 'mpd',
        'f4m': 'f4m',
        'f4m+xml': 'f4m',
        'hds+xml': 'f4m',
        'vnd.ms-sstr+xml': 'ism',
    }
    return ext[res] || res

}


/*** Extraction process ***/

function YoutubeExtractor(){
    this._ciphers = {}
}
YoutubeExtractor.prototype = Object.create(require('events').EventEmitter.prototype)

YoutubeExtractor.prototype.notifyProgress = function(dl_id, data){
    var self = this
    data = data || {}
    if(dl_id){
        data.id = dl_id
        self.emit('progress', data)
    }
}

YoutubeExtractor.prototype.notifyError = function(dl_id, e, url){
    var self = this
    if(e.message)
        e = e.message
    var data = {
        message: e,
        id: dl_id,
        url: url
    }
    self.emit('error', data)
}

YoutubeExtractor.prototype.real_extract = function(url, dl_id){
    var self = this

    var youtube_include_dash_manifest = true

    /*** All variables returned by this function ***/
    var video_info;
    var video_id;
    var video_uploader;
    var video_uploader_id;
    var video_uploader_url;
    var upload_date;
    var video_license;
    var video_creator;
    var video_title;
    var video_alt_title;
    var video_thumbnail;
    var video_description;
    var video_categories;
    var video_tags;
    var video_subtitles;
    var automatic_captions;
    var video_duration;
    var age_gate;
    var video_annotations;
    var view_count;
    var like_count;
    var dislike_count;
    var formats;
    var is_live;
    var start_time;
    var end_time;

    //Format list
    var _formats = {
        '5': {'ext': 'flv', 'width': 400, 'height': 240, 'acodec': 'mp3', 'abr': 64, 'vcodec': 'h263'},
        '6': {'ext': 'flv', 'width': 450, 'height': 270, 'acodec': 'mp3', 'abr': 64, 'vcodec': 'h263'},
        '13': {'ext': '3gp', 'acodec': 'aac', 'vcodec': 'mp4v'},
        '17': {'ext': '3gp', 'width': 176, 'height': 144, 'acodec': 'aac', 'abr': 24, 'vcodec': 'mp4v'},
        '18': {'ext': 'mp4', 'width': 640, 'height': 360, 'acodec': 'aac', 'abr': 96, 'vcodec': 'h264'},
        '22': {'ext': 'mp4', 'width': 1280, 'height': 720, 'acodec': 'aac', 'abr': 192, 'vcodec': 'h264'},
        '34': {'ext': 'flv', 'width': 640, 'height': 360, 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264'},
        '35': {'ext': 'flv', 'width': 854, 'height': 480, 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264'},
        //itag 36 videos are either 320x180 (BaW_jenozKc) or 320x240 (__2ABJjxzNo), abr varies as well
        '36': {'ext': '3gp', 'width': 320, 'acodec': 'aac', 'vcodec': 'mp4v'},
        '37': {'ext': 'mp4', 'width': 1920, 'height': 1080, 'acodec': 'aac', 'abr': 192, 'vcodec': 'h264'},
        '38': {'ext': 'mp4', 'width': 4096, 'height': 3072, 'acodec': 'aac', 'abr': 192, 'vcodec': 'h264'},
        '43': {'ext': 'webm', 'width': 640, 'height': 360, 'acodec': 'vorbis', 'abr': 128, 'vcodec': 'vp8'},
        '44': {'ext': 'webm', 'width': 854, 'height': 480, 'acodec': 'vorbis', 'abr': 128, 'vcodec': 'vp8'},
        '45': {'ext': 'webm', 'width': 1280, 'height': 720, 'acodec': 'vorbis', 'abr': 192, 'vcodec': 'vp8'},
        '46': {'ext': 'webm', 'width': 1920, 'height': 1080, 'acodec': 'vorbis', 'abr': 192, 'vcodec': 'vp8'},
        '59': {'ext': 'mp4', 'width': 854, 'height': 480, 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264'},
        '78': {'ext': 'mp4', 'width': 854, 'height': 480, 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264'},

        //3D videos
        '82': {'ext': 'mp4', 'height': 360, 'format_note': '3D', 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264', 'preference': -20},
        '83': {'ext': 'mp4', 'height': 480, 'format_note': '3D', 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264', 'preference': -20},
        '84': {'ext': 'mp4', 'height': 720, 'format_note': '3D', 'acodec': 'aac', 'abr': 192, 'vcodec': 'h264', 'preference': -20},
        '85': {'ext': 'mp4', 'height': 1080, 'format_note': '3D', 'acodec': 'aac', 'abr': 192, 'vcodec': 'h264', 'preference': -20},
        '100': {'ext': 'webm', 'height': 360, 'format_note': '3D', 'acodec': 'vorbis', 'abr': 128, 'vcodec': 'vp8', 'preference': -20},
        '101': {'ext': 'webm', 'height': 480, 'format_note': '3D', 'acodec': 'vorbis', 'abr': 192, 'vcodec': 'vp8', 'preference': -20},
        '102': {'ext': 'webm', 'height': 720, 'format_note': '3D', 'acodec': 'vorbis', 'abr': 192, 'vcodec': 'vp8', 'preference': -20},

        //Apple HTTP Live Streaming
        '91': {'ext': 'mp4', 'height': 144, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 48, 'vcodec': 'h264', 'preference': -10},
        '92': {'ext': 'mp4', 'height': 240, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 48, 'vcodec': 'h264', 'preference': -10},
        '93': {'ext': 'mp4', 'height': 360, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264', 'preference': -10},
        '94': {'ext': 'mp4', 'height': 480, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 128, 'vcodec': 'h264', 'preference': -10},
        '95': {'ext': 'mp4', 'height': 720, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 256, 'vcodec': 'h264', 'preference': -10},
        '96': {'ext': 'mp4', 'height': 1080, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 256, 'vcodec': 'h264', 'preference': -10},
        '132': {'ext': 'mp4', 'height': 240, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 48, 'vcodec': 'h264', 'preference': -10},
        '151': {'ext': 'mp4', 'height': 72, 'format_note': 'HLS', 'acodec': 'aac', 'abr': 24, 'vcodec': 'h264', 'preference': -10},

        //DASH mp4 video
        '133': {'ext': 'mp4', 'height': 240, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},
        '134': {'ext': 'mp4', 'height': 360, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},
        '135': {'ext': 'mp4', 'height': 480, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},
        '136': {'ext': 'mp4', 'height': 720, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},
        '137': {'ext': 'mp4', 'height': 1080, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},
        '138': {'ext': 'mp4', 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},  // Height can vary (https://github.com/rg3/youtube-dl/issues/4559)
        '160': {'ext': 'mp4', 'height': 144, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},
        '264': {'ext': 'mp4', 'height': 1440, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},
        '298': {'ext': 'mp4', 'height': 720, 'format_note': 'DASH video', 'vcodec': 'h264', 'fps': 60, 'preference': -40},
        '299': {'ext': 'mp4', 'height': 1080, 'format_note': 'DASH video', 'vcodec': 'h264', 'fps': 60, 'preference': -40},
        '266': {'ext': 'mp4', 'height': 2160, 'format_note': 'DASH video', 'vcodec': 'h264', 'preference': -40},

        //Dash mp4 audio
        '139': {'ext': 'm4a', 'format_note': 'DASH audio', 'acodec': 'aac', 'abr': 48, 'preference': -50, 'container': 'm4a_dash'},
        '140': {'ext': 'm4a', 'format_note': 'DASH audio', 'acodec': 'aac', 'abr': 128, 'preference': -50, 'container': 'm4a_dash'},
        '141': {'ext': 'm4a', 'format_note': 'DASH audio', 'acodec': 'aac', 'abr': 256, 'preference': -50, 'container': 'm4a_dash'},
        '256': {'ext': 'm4a', 'format_note': 'DASH audio', 'acodec': 'aac', 'preference': -50, 'container': 'm4a_dash'},
        '258': {'ext': 'm4a', 'format_note': 'DASH audio', 'acodec': 'aac', 'preference': -50, 'container': 'm4a_dash'},

        //Dash webm
        '167': {'ext': 'webm', 'height': 360, 'width': 640, 'format_note': 'DASH video', 'container': 'webm', 'vcodec': 'vp8', 'preference': -40},
        '168': {'ext': 'webm', 'height': 480, 'width': 854, 'format_note': 'DASH video', 'container': 'webm', 'vcodec': 'vp8', 'preference': -40},
        '169': {'ext': 'webm', 'height': 720, 'width': 1280, 'format_note': 'DASH video', 'container': 'webm', 'vcodec': 'vp8', 'preference': -40},
        '170': {'ext': 'webm', 'height': 1080, 'width': 1920, 'format_note': 'DASH video', 'container': 'webm', 'vcodec': 'vp8', 'preference': -40},
        '218': {'ext': 'webm', 'height': 480, 'width': 854, 'format_note': 'DASH video', 'container': 'webm', 'vcodec': 'vp8', 'preference': -40},
        '219': {'ext': 'webm', 'height': 480, 'width': 854, 'format_note': 'DASH video', 'container': 'webm', 'vcodec': 'vp8', 'preference': -40},
        '278': {'ext': 'webm', 'height': 144, 'format_note': 'DASH video', 'container': 'webm', 'vcodec': 'vp9', 'preference': -40},
        '242': {'ext': 'webm', 'height': 240, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '243': {'ext': 'webm', 'height': 360, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '244': {'ext': 'webm', 'height': 480, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '245': {'ext': 'webm', 'height': 480, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '246': {'ext': 'webm', 'height': 480, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '247': {'ext': 'webm', 'height': 720, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '248': {'ext': 'webm', 'height': 1080, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '271': {'ext': 'webm', 'height': 1440, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        //itag 272 videos are either 3840x2160 (e.g. RtoitU2A-3E) or 7680x4320 (sLprVF6d7Ug)
        '272': {'ext': 'webm', 'height': 2160, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '302': {'ext': 'webm', 'height': 720, 'format_note': 'DASH video', 'vcodec': 'vp9', 'fps': 60, 'preference': -40},
        '303': {'ext': 'webm', 'height': 1080, 'format_note': 'DASH video', 'vcodec': 'vp9', 'fps': 60, 'preference': -40},
        '308': {'ext': 'webm', 'height': 1440, 'format_note': 'DASH video', 'vcodec': 'vp9', 'fps': 60, 'preference': -40},
        '313': {'ext': 'webm', 'height': 2160, 'format_note': 'DASH video', 'vcodec': 'vp9', 'preference': -40},
        '315': {'ext': 'webm', 'height': 2160, 'format_note': 'DASH video', 'vcodec': 'vp9', 'fps': 60, 'preference': -40},

        //Dash webm audio
        '171': {'ext': 'webm', 'acodec': 'vorbis', 'format_note': 'DASH audio', 'abr': 128, 'preference': -50},
        '172': {'ext': 'webm', 'acodec': 'vorbis', 'format_note': 'DASH audio', 'abr': 256, 'preference': -50},

        //Dash webm audio with opus inside
        '249': {'ext': 'webm', 'format_note': 'DASH audio', 'acodec': 'opus', 'abr': 50, 'preference': -50},
        '250': {'ext': 'webm', 'format_note': 'DASH audio', 'acodec': 'opus', 'abr': 70, 'preference': -50},
        '251': {'ext': 'webm', 'format_note': 'DASH audio', 'acodec': 'opus', 'abr': 160, 'preference': -50},

        //RTMP (unnamed)
        '_rtmp': {'protocol': 'rtmp'}
    }

    //Youtube regex
    var re = /(?:youtube\.com\/\S*(?:(?:\/e(?:mbed))?\/|watch\/?\?(?:\S*?&?v=))|youtu\.be\/)([a-zA-Z0-9_-]{6,11})/g;
    //Temporary regex execution result variable
    var mobj;

    //Globals
    var parsed_url = Url.parse(url, true);
    var player_url;

    var video_webpage;
    var embed_webpage;

    var video_info_url;
    var video_info_webpage;


    //Dash mpds handling (??)
    var dash_mpds;
    function add_dash_mpd (video_info){
        var dash_mpd = video_info['dashmpd']
        if(dash_mpd && dash_mpds.indexOf(dash_mpd[0])==-1)
            dash_mpds.push(dash_mpd[0])
    }

    //Returning a promises chain
    return Q().then(function(){
        /*** Parsing url parameters and downloading youtube video webpage ***/
        self.notifyProgress(dl_id, {
            step: 'Finding video id in url'
        })

        //Parsing start / end times if present in url query
        for(var query in parsed_url['query']){
            if(!start_time && query=="t")
                start_time = new Date(parsed_url.query[query]);
            if(!start_time && query=="start")
                start_time = new Date(parsed_url.query[query]);
            if(!end_time && query=="end")
                start_time = new Date(parsed_url.query[query]);
        }

        //Fetching video ID
        mobj = re.exec(url)
        if(!mobj || !mobj.length || ! mobj[1])
            throw "Video ID not found !"
        video_id = mobj[1]

        //Rebuild url
        url = "https://www.youtube.com/watch?v=" + video_id + "&gl=US&hl=en&has_verified=1&bpctr=9999999999"

        //Downloading youtube webpage
        return self._download_webpage(url, "Downloading Youtube webpage", dl_id)

    }).then(function(res){

        //Assign request result to the 'video_webpage' function global
        video_webpage = res

        //Getting player url from swf config if present
        if(/swfConfig.*?"(https?:\\\/\\\/.*?watch.*?-.*?\.swf)"/.exec(video_webpage)){
            //TODO swf handling
        }
        else{
            player_url = null
        }

    }).then(function(){
        /*** Handling asynchronous getting of info if age restriction ***/

        self.notifyProgress(dl_id, {
            step: 'Finding video info in HTML page'
        })
        //Get video info
        dash_mpds = []
        embed_webpage = null
        is_live = null

        if(/player-age-gate-content">/.exec(video_webpage)){
            age_gate = true

            url = "https://www.youtube.com/embed/" + video_id

            return self._download_webpage(url, 'Downloading embed webpage', dl_id).then(function(res){
                embed_webpage = res

                var re_res = /"sts"\s*:\s*(\d+)/.exec(embed_webpage)
                var data = {
                    'video_id': video_id,
                    'eurl': 'https://youtube.googleapis.com/v/' + video_id,
                    'sts': re_res? re_res[1] : ''
                }
                video_info_url = _build_url("https://www.youtube.com/get_video_info", data)
                return self._download_webpage(video_info_url, "Refetching age-gated info webpage", dl_id).then(function(res){
                    video_info_webpage = res
                    video_info = _parse_video_info(_parse_query(video_info_webpage))
                    add_dash_mpd(video_info)
                })
            })
        }
        else{
            age_gate = false
            video_info = null

            var conf = /;ytplayer\.config\s*=\s*({.+?});/.exec(video_webpage)
            if(!conf)
                throw "No player config !"
            var ytplayer_config = JSON.parse(conf[1])
            if(ytplayer_config){
                var args = ytplayer_config['args']
                if(args.hasOwnProperty('url_encoded_fmt_stream_map')){
                    video_info = _parse_video_info(args)
                    add_dash_mpd(video_info)
                }
                is_live = args['livestream'] == '1' || args['live_playback'] == 1
            }
            if(!video_info || youtube_include_dash_manifest){
                var promises = [ ]
                var should_exec = true
                var elts = ['&el=info', '&el=embedded', '&el=detailpage', '&el=vevo', '']
                elts.forEach(function(el_type){
                    video_info_url = "https://www.youtube.com/get_video_info?&video_id=" + video_id + el_type + "&ps=default&eurl=&gl=US&hl=en"
                    var promise =  self._download_webpage(video_info_url, "Refetching video info").then(function(res){
                        if(should_exec){
                            video_info_webpage = res
                            var get_video_info = _parse_video_info(_parse_query(video_info_webpage))
                            if(!!get_video_info['use_cipher_signature'] && get_video_info['use_cipher_signature'][0]!='True')
                                add_dash_mpd(get_video_info)
                            if(!video_info)
                                video_info = get_video_info
                            if(get_video_info['token']){
                                if(!video_info['token'])
                                    video_info = get_video_info
                                should_exec = false
                            }
                        }
                    })

                    promises.push(promise)
                })
                return Q.all(promises)
            }
        }

    }).then(function() {
        //console.log(JSON.stringify(video_info))
        self.notifyProgress(dl_id, {
            step: 'Extracting metadata from video info'
        })
        // Main info extraction process

        //Creating a jquery-like object for utility functions
        var $ = cheerio.load(video_webpage)

        //Token
        if (!video_info['token'])
            throw video_info['reason'] || 'No token !'

        //Title
        if (video_info['title'])
            video_title = video_info['title'][0] ? decodeURI(video_info['title'][0]).replace(/\+/g, ' ') : '_'

        //Video description
        var desc_elt = $('#eow-description')
        desc_elt.html(desc_elt.html().replace(/<br>/g, '{{line}}'))
        video_description = desc_elt.text().replace(/\{\{line\}\}/g, '\n')
        if (!video_description) {
            var searched = /<meta name="description" content="([^"]+)"/.exec(video_webpage)
            video_description = searched ? searched[1] : ''
        }

        //View count
        if (video_info['view_count'])
            view_count = parseInt(video_info['view_count'][0])

        //Rental video check
        if (video_info['ypc_video_rental_bar_text'] && !video_info['author'])
            throw "Rental videos not supported"

        //Author - Uploader info
        if (!video_info['author'])
            throw "No author !"
        video_uploader = decodeURI(video_info['author'][0])
        mobj = /<link itemprop="url" href="(https?:\/\/www.youtube.com\/(?:user|channel)\/([^"]+))">/.exec(video_webpage)
        if (mobj) {
            video_uploader_id = mobj[2]
            video_uploader_url = mobj[1]
        }

        //Thumbnail image
        var m_thumb = /<span itemprop="thumbnail".*?href="(.*?)">/gm.exec(video_webpage)
        if (m_thumb)
            video_thumbnail = m_thumb[1]
        else if (video_info['thumbnail_url'])
            video_thumbnail = decodeURI(video_info['thumbnail_url'])

        //Upload date
        var date_meta = search_meta($, 'datePublished')
        if (date_meta) {
            upload_date = date_meta.attr('content')
        }
        else {
            //Alternative date fetching
            var res = [/id="eow-date.*?>(.*?)<\/span>/, /id="watch-uploader-info".*?>.*?(?:Published|Uploaded|Streamed live|Started) on (.+?)<\/strong>/]
            res.forEach(function (re) {
                mobj = re.exec(video_webpage)
                if (mobj && mobj[1]) {
                    upload_date = mobj[1]
                }
            })
        }
        upload_date = new Date(upload_date)

        //Video license
        mobj = /<h4[^>]+class="title"[^>]*>\s*License\s*<\/h4>\s*<ul[^>]*>\s*<li>(.+?)<\/li/.exec(video_webpage)
        video_license = mobj ? mobj[1] : ''

        //Alt title / creator (music)
        var m_music = /<h4[^>]+class="title"[^>]*>\s*Music\s*<\/h4>\s*<ul[^>]*>\s*<li>(.+?) by (.+?)(?:\(.+?\))?<\/li/.exec(video_webpage)
        if (m_music) {
            video_alt_title = htmlToText.fromString(m_music[1])
            video_alt_title = video_alt_title.substring(1, video_alt_title.length - 1)
            video_creator = htmlToText.fromString(m_music[2])
        }

        //Categories
        var cat_meta = search_meta($, 'genre')
        if (cat_meta) {
            video_categories = [cat_meta.attr('content')]
        }

        //Video tags
        var tags = search_meta($, 'og:video:tag')
        video_tags = []
        tags.each(function () {
            video_tags.push($(this).attr('content'))
        })

        //Like / dislike counts
        var m_count = /-like-button[^>]+><span[^>]+class="yt-uix-button-content"[^>]*>([\d,]+)<\/span>/.exec(video_webpage)
        if (m_count)
            like_count = parseInt(m_count[1].replace(/,/g, ''))
        m_count = /-dislike-button[^>]+><span[^>]+class="yt-uix-button-content"[^>]*>([\d,]+)<\/span>/.exec(video_webpage)
        if (m_count)
            dislike_count = parseInt(m_count[1].replace(/,/g, ''))

        //We don't process subtitles / captions / annotations

        //Video duration
        if (video_info['length_seconds'])
            video_duration = parseInt(decodeURI(video_info['length_seconds'][0]))

    }).then(function() {
        var promises = []
        //FORMATS HANDLING
        self.notifyProgress(dl_id, {
            step: 'Fetching formats'
        })

        //Function to handle hlsvp urlmap
        function map_to_format_list(urlmap) {
            var formats = []
            for (var key in urlmap) {
                var dct = {
                    'format_id': key,
                    'url': urlmap[key],
                    'player_url': player_url
                }
                if (_formats[key])
                    _.merge(dct, _formats[key])
                formats.push(dct)
            }
            return formats
        }

        //Special format rtmp ??
        if (video_info['conn'] && video_info['conn'][0].startsWith('rtmp')) {
            formats = [{
                'format_id': '_rtmp',
                'protocol': 'rtmp',
                'url': video_info['conn'][0],
                'player_url': player_url
            }]
        }
        //Typical format (for test video)
        else if (video_info['url_encoded_fmt_stream_map'] && video_info['url_encoded_fmt_stream_map'][0].length >= 1 || video_info['adaptive_fmts'] && video_info['adaptive_fmts'][0].length >= 1) {
            var encoded_url_map = (video_info['url_encoded_fmt_stream_map'] ? video_info['url_encoded_fmt_stream_map'][0] : '') + ',' + (video_info['adaptive_fmts'] ? video_info['adaptive_fmts'][0] : '')
            if (encoded_url_map.indexOf('rtmpe%3Dyes') != -1)
                throw 'rtmpe downloads are not supported, see https://github.com/rg3/youtube-dl/issues/343 for more information.'
            var formats_spec = {}
            var fmt_list = video_info['fmt_list'] ? video_info['fmt_list'][0] : ''
            if (fmt_list) {
                fmt_list.split(',').forEach(function (fmt) {
                    var spec = fmt.split('/')
                    if (spec && spec.length > 1) {
                        var width_height = spec[1].split('x')
                        if (width_height.length == 2) {
                            formats_spec[spec[0]] = {
                                'resolution': spec[1],
                                'width': width_height[0] ? parseInt(width_height[0]) : null,
                                'height': width_height[1] ? parseInt(width_height[1]) : null
                            }
                        }
                    }
                })
            }
            formats = []
            encoded_url_map.split(',').forEach(function (url_data_str) {
                var url_data = _parse_video_info(_parse_query(url_data_str))
                if (!url_data['itag'] || !url_data['url'])
                    return
                var format_id = url_data['itag'][0]
                var url = url_data['url'][0]


                function _after(){
                    if (url.indexOf('ratebypass')==-1)
                        url += '&ratebypass=yes'

                    var dct = {
                        'format_id': format_id,
                        'url': url,
                        'player_url': player_url
                    }

                    _.merge(dct, _formats[format_id], formats_spec[format_id])

                    mobj = /^(\d+)[xX](\d+)$/.exec(url_data['size'] ? url_data['size'][0] : '')
                    var width = mobj ? mobj[1] : null,
                        height = mobj ? mobj[2] : null
                    var more_fields = {
                        'filesize': url_data['clen'] ? parseInt(url_data['clen'][0]) : null,
                        'tbr': url_data['bitrate'] ? parseFloat(url_data['bitrate'][0])/1000 : null,
                        'width': width,
                        'height': height,
                        'fps': url_data['fps'] ? parseInt(url_data['fps'][0]) : null,
                        'format_note': url_data['quality_label'] ? url_data['quality_label'][0] : (url_data['quality'] ? url_data['quality'][0] : null)
                    }
                    for (var key in more_fields) {
                        if (more_fields[key]) {
                            dct[key] = more_fields[key]
                        }
                    }
                    var type_ = url_data['type'] ? url_data['type'][0] : null
                    if (type_) {
                        var type_split = type_.split(';')
                        var kind_ext = type_split[0].split('/')
                        if (kind_ext.length == 2) {
                            var kind = kind_ext[0]
                            dct['ext'] = mimetype2ext(type_split[0])
                            if (['audio', 'video'].indexOf(kind) != -1) {
                                var codecs;
                                var mobj = /codecs=["']?(.+?)["']?(?:;|$)/.exec(type_)
                                if (mobj)
                                    codecs = mobj[1]

                                if (codecs) {
                                    codecs = codecs.split(',')
                                    if (codecs.length == 2)
                                        _.merge(dct, {'acodec': decodeURI(codecs[1]).replace(/\+/g, ' '), 'vcodec': decodeURI(codecs[0]).replace(/\+/g, ' ')})
                                    else
                                        _.merge(dct, {
                                            'acodec': kind == 'audio' ? decodeURI(codecs[0]).replace(/\+/g, ' ') : null,
                                            'vcodec': kind == 'audio' ? null : decodeURI(codecs[0]).replace(/\+/g, ' ')
                                        })
                                }
                            }
                        }
                    }
                    formats.push(dct)
                }


                if (url_data['sig']) {
                    url += '&signature=' + url_data['sig'][0]
                    _after()
                }
                else if (url_data['s']) {
                    var encrypted_sig = url_data['s'][0]
                    var ASSETS_RE = /"assets":.+?"js":\s*("[^"]+")/
                    var jsplayer_url_json = /"assets":.+?"js":\s*("[^"]+")/.exec(age_gate ? embed_webpage : video_webpage)
                    if (!jsplayer_url_json && !age_gate) {
                        //TODO
                    }
                    player_url = jsplayer_url_json ? JSON.parse(jsplayer_url_json[1]) : null
                    if (!player_url) {
                        jsplayer_url_json = /ytplayer\.config.*?"url"\s*:\s*("[^"]+")/.exec(video_webpage)
                        player_url = jsplayer_url_json ? JSON.parse(jsplayer_url_json[1]) : null
                    }

                    promises.push(function(){
                        return self.decrypt_signature(player_url, encrypted_sig).then(function(signature){
                            url += '&signature=' + signature
                            _after()
                        })
                    })

                }
                else{
                    _after()
                }
            })

        }
        else if (video_info['hlsvp']) {
            var manifest_url = video_info['hlsvp'][0]
            return self.extract_from_m3u8(manifest_url, dl_id).then(function (url_map) {
                formats = map_to_format_list(url_map)
                formats.forEach(function (a_format) {
                    if (!a_format['http_headers'])
                        a_format['http_headers'] = {}
                    a_format['http_headers']['Youtubedl-no-compression'] = true
                })
            })
        }
        else {
            throw "Video unavailable !"
        }

        return promises.reduce(Q.when, Q());
    }).then(function(){
        //Look for the dash manifest


    }).then(function(){
        /*** Return extracted info  ***/
        return {
            'id': video_id,
            'uploader': video_uploader,
            'uploader_id': video_uploader_id,
            'uploader_url': video_uploader_url,
            'upload_date': upload_date,
            'license': video_license,
            'creator': video_creator,
            'title': video_title,
            'alt_title': video_alt_title,
            'thumbnail': video_thumbnail,
            'description': video_description,
            'categories': video_categories,
            'tags': video_tags,
            'subtitles': video_subtitles,
            'automatic_captions': automatic_captions,
            'duration': video_duration,
            'age_limit': age_gate? 18 : 0,
            'annotations': video_annotations,
            'webpage_url': 'https://www.youtube.com/watch?v=' + video_id,
            'view_count': view_count,
            'like_count': like_count,
            'dislike_count': dislike_count,
            'average_rating': video_info['avg_rating'] || 0,
            'formats': formats,
            'is_live': !!is_live,
            'start_time': start_time,
            'end_time': end_time
        }
    })

}

//Player signature decrypt
YoutubeExtractor.prototype.decrypt_signature = function(player_url, s){
    var self = this
    if(!player_url)
        throw "No player url for signature decrypt!"
    if(player_url.startsWith('//'))
        player_url = "https:" + player_url
    try{
        var id_m = /.*?-([a-zA-Z0-9_-]+)(?:\/watch_as3|\/html5player(?:-new)?|\/base)?\.([a-z]+)$/.exec(player_url)
        var player_type = id_m[2]
        var player_id = id_m[1]

        if(player_type=='js'){
            if(!self._ciphers[player_id]) {
                self._ciphers[player_id] = {
                    cipher: new CipherExtract(player_url)
                }
            }
            return self._ciphers[player_id]['cipher'].getSigFunction().then(function(fn){
                return fn(s)
            })
        }
        else if(player_type=='swf'){
            //TODO swf player parsing
            return Q.when(s)
        }
        else{
            throw "Invalid player type"
        }
    }
    catch(e){
        throw "Signature extraction failed : "+e
    }
}

YoutubeExtractor.prototype.extract_from_m3u8 = function(manifest_url, dl_id){
    var self = this

    var url_map = {}

    function _get_urls(_manifest){
        var lines = _manifest.split('\n')
        var urls = []
        lines.forEach(function(l){
            l && !l.startsWith('#') && urls.push(l)
        })
        return urls
    }

    return Q().then(function(){
        return self._download_webpage(manifest_url, "Downloading formats manifest", dl_id)
    }).then(function(manifest){
        var formats_urls = _get_urls(manifest)
        formats_urls.forEach(function(format_url){
            var m_itag = /itag\/(\d+?)\//.exec(format_url)
            if(m_itag)
                url_map[m_itag[1]] = format_url
        })
        return url_map
    })
}

YoutubeExtractor.prototype._download_webpage = function(url, msg, dl_id){
    var self = this
    self.notifyProgress(dl_id, {
        step: msg
    })
    var deferred = Q.defer();

    var parsed_url = Url.parse(url)
    var options = {
        "method": "GET",
        "hostname": parsed_url.hostname,
        "port": null,
        "path": '/' + parsed_url.path
    };

    var req = http.request(options, function (res) {
        var chunks = [];

        res.on("data", function (chunk) {
            chunks.push(chunk);
        });

        res.on("end", function () {
            var body = Buffer.concat(chunks);
            deferred.resolve(body.toString());
        });
    });

    req.on('error', function(err){
        deferred.reject(err);
    });

    req.end();

    return deferred.promise;
}

YoutubeExtractor.prototype._download_file = function(url, filename, dl_id){
    var self = this
    self.notifyProgress(dl_id, {
        step: 'Downloading '+filename
    })
    var log_locker = false
    var deferred = Q.defer();

    var parsed_url = Url.parse(url)
    var options = {
        "method": "GET",
        "hostname": parsed_url.hostname,
        "port": null,
        "path": '/' + parsed_url.path
    };

    var http
    if(parsed_url.protocol.indexOf('https')!=-1)
        http = require('https')
    else
        http = require('http')

    var file = fs.createWriteStream(filename)

    var req = http.request(options, function (res) {
        if(res.statusCode==302 && res.headers.location){
            self._download_file(res.headers.location, filename, dl_id).then(deferred.resolve, deferred.reject)
        }
        else{
            var progress = 0
            var length = 0
            res.on('data', function(data) {
                length += data.length
                if(res.headers['content-length']){
                    progress = length / parseInt(res.headers['content-length'])
                }
                if(!log_locker){
                    log_locker = true
                    setTimeout(function(){
                        log_locker = false
                    }, 500)
                    self.notifyProgress(dl_id, {
                        step: 'Downloading '+filename,
                        progress: progress
                    })
                }
                file.write(data);
            }).on('end', function() {
                file.end();
                console.log(filename, 'downloaded');
                deferred.resolve()
            })
        }
    })

    req.on('error', function(err){
        deferred.reject(err);
    });

    req.end();

    return deferred.promise;

}

YoutubeExtractor.prototype._delete_file = function(filename, dl_id){
    var deferred = Q.defer()
    fs.unlink(filename, function(err){
        if(err)
            deferred.reject(err)
        else
            deferred.resolve()
    })
    return deferred.promise
}

YoutubeExtractor.prototype.download_video = function(ieResult, path, dl_id){
    var self = this
    self.notifyProgress(dl_id, {
        step: 'Downloading video'
    })

    var max_br = 0
    var format = {}
    var fileData = {}
    ieResult.formats.forEach(function(elt){
        if(elt['format_note']=='DASH audio' && elt['url']){
            if(elt['abr']>max_br && elt['ext']=='webm'){
                max_br = elt['abr']
                format = elt
            }
        }
    })
    fileData.filename = ieResult.title.replace(/\//g, '')
    if(!format.url){
        max_br = 0
        ieResult.formats.forEach(function(elt){
            if(elt['abr']>max_br){
                max_br = elt['abr']
                format = elt
            }
        })
    }
    if(!format.url)
        throw 'No suitable format found'
    return self._download_file(format.url, Path.join(path, fileData.filename+'.'+format.ext), dl_id).then(function(){
        var filename = fileData.filename
        return {
            path: path,
            filename: filename,
            ext: format.ext,
            abr: format.abr || 128
        }
    })
}

YoutubeExtractor.prototype.postprocess = function(data, dl_id){
    var self = this
    self.notifyProgress(dl_id, {
        step: 'Postprocessing into mp3'
    })

    var deferred = Q.defer()
    var video_filename = Path.join(data.path, data.filename+'.'+data.ext)
    var audio_filename = Path.join(data.path, data.filename+'.mp3')
    ffmpeg(video_filename)
        .audioCodec('libmp3lame')
        .audioBitrate(data.abr)
        .on('progress', function(progress) {
            self.notifyProgress(dl_id, {
                step: 'Postprocessing into mp3',
                progress: (progress && progress.percent)? (progress.percent/100):0
            })
        })
        .on('end', function(){
            deferred.resolve()
        })
        .on('error', function(err){
            deferred.reject(err)
        })
        .save(audio_filename)
    return deferred.promise
}

YoutubeExtractor.prototype.downloadMp3 = function(url, path){
    var self = this
    var fileData = {}

    path = path || ''

    var dl_id = Date.now()

    self.notifyProgress(dl_id, {
        step: 'Starting extraction process'
    })

    Q()
        .then(function() {
            return self.real_extract(url, dl_id)
        }).then(function(res){
            return self.download_video(res, path, dl_id)
        }).then(function(data){
            fileData = data
            return self.postprocess(fileData, dl_id)
        }).then(function(){
            return self._delete_file(Path.join(path, fileData.filename+'.'+fileData.ext), dl_id)
        }).then(function(){
            var file = Path.join(path, fileData.filename+'.mp3')
            self.emit('end', {
                id: dl_id,
                title: fileData.filename,
                file: file,
                url: url
            })
            return file
        }).catch(function(e){
            self.notifyError(dl_id, e, url)
        })

    return dl_id
}

module.exports = YoutubeExtractor